#!/bin/bash

######################################################################
# Proprietary Information D-Wave Systems Inc.                        #
# Copyright (c) 2017 by D-Wave Systems Inc. All rights reserved.     #
# D-Wave Systems Inc., 3033 Beta Ave., Burnaby, BC, V5G 4M9, Canada. #
######################################################################

usage="dw-get-object.bash {Qdddd|Cdddd}"

if [[ $# != 1 || ( $1 != Q[0-9][0-9][0-9][0-9] &&  $1 != C[0-9][0-9][0-9][0-9] ) ]]
then
    echo "$usage"
    exit 1
fi

awk -v selector=$1 '

/^L:/ { DW_L = $2 }
/^M:/ { DW_M = $2 }
/^N:/ { DW_N = $2 }

/^Q/ { gsub(" ", "");
       split($0, token, ":");
       qubit_range = token[1];
       bit_mask = token[2];
       sub("-.*$", "", qubit_range);
       sub("^Q", "", qubit_range);
       qubit_index = 0 + qubit_range;
       bit_index = 1;
       bits = length(bit_mask);

       while (bit_index <= bits)
       {
           qubit[qubit_index] = substr(bit_mask, bit_index, 1);
           ++ qubit_index;
           ++ bit_index;
       }
}

/^C/ { gsub(" ", "");
       split($0, token, ":");
       coupler_range = token[1];
       bit_mask = token[2];
       sub("-.*$", "", coupler_range);
       sub("^C", "", coupler_range);
       coupler_index = 0 + coupler_range;
       bit_index = 1;
       bits = length(bit_mask);

       while (bit_index <= bits)
       {
           coupler[coupler_index] = substr(bit_mask, bit_index, 1);
           ++ coupler_index;
           ++ bit_index;
       }
}


END {
    qubits = 2*DW_L*DW_M*DW_N;

    couplers = DW_L*DW_L*DW_M*DW_N + DW_L*DW_M*(DW_N-1) + DW_L*(DW_M-1)*DW_N;

    if (selector ~ /Q[0-9][0-9][0-9][0-9]/)
    {
        qnum = selector;
        sub("Q", "", qnum);
        qnum = 0 + qnum;

        if (qnum >= qubits)
        {
            printf "error: %s is out of range for geometry\n", selector;
            exit 1;
        }

        # Info about Qdddd:
        ChimInd_from_Linear(qnum);
        qnum_i = i;
        qnum_j = j;
        qnum_u = u;
        qnum_k = k;
        printf "Q%04d|%s|ChimeraIndex(%d,%d,%d,%d)\n",
            qnum, (0 + qubit[qnum]) ? "active" : "inactive",
            qnum_i, qnum_j, qnum_u, qnum_k;

        # Intracell neighbors:
        for (n=0; n<DW_L; ++n)
        {
            qneighbor = Linear_from_ChimInd(qnum_i, qnum_j, 1-qnum_u, n);
            cnum = Coupler_from_Qubits(qnum, qneighbor);
            state = (0 + qubit[qneighbor]) ? "active" : "inactive";
            printf "Q%04d <== C%04d|%s|%s ==> Q%04d|%s|ChimeraIndex(%d,%d,%d,%d)\n",
                qnum,
                cnum, (0 + coupler[cnum]) ? "active" : "inactive", CouplerType(cnum),
                qneighbor, state, qnum_i, qnum_j, 1-qnum_u, n;
        }

        # Low horizontal neighbor:
        if (qnum_j > 0 && qnum_u == 1)
        {
            qneighbor = Linear_from_ChimInd(qnum_i, -1+qnum_j, qnum_u, qnum_k);
            cnum = Coupler_from_Qubits(qnum, qneighbor);
            state = (0 + qubit[qneighbor]) ? "active" : "inactive";
            printf "Q%04d <== C%04d|%s|%s ==> Q%04d|%s|ChimeraIndex(%d,%d,%d,%d)\n",
                qnum,
                cnum, (0 + coupler[cnum]) ? "active" : "inactive", CouplerType(cnum),
                qneighbor, state, qnum_i, -1+qnum_j, qnum_u, qnum_k;
        }

        # High horizontal neighbor:
        if (qnum_j < (-1+DW_N) && qnum_u == 1)
        {
            qneighbor = Linear_from_ChimInd(qnum_i, +1+qnum_j, qnum_u, qnum_k);
            cnum = Coupler_from_Qubits(qnum, qneighbor);
            state = (0 + qubit[qneighbor]) ? "active" : "inactive";
            printf "Q%04d <== C%04d|%s|%s ==> Q%04d|%s|ChimeraIndex(%d,%d,%d,%d)\n",
                qnum,
                cnum, (0 + coupler[cnum]) ? "active" : "inactive", CouplerType(cnum),
                qneighbor, state, qnum_i, +1+qnum_j, qnum_u, qnum_k;
        }

        # Low vertical neighbor:
        if (qnum_i > 0 && qnum_u == 0)
        {
            qneighbor = Linear_from_ChimInd(-1+qnum_i, qnum_j, qnum_u, qnum_k);
            cnum = Coupler_from_Qubits(qnum, qneighbor);
            state = (0 + qubit[qneighbor]) ? "active" : "inactive";
            printf "Q%04d <== C%04d|%s|%s ==> Q%04d|%s|ChimeraIndex(%d,%d,%d,%d)\n",
                qnum,
                cnum, (0 + coupler[cnum]) ? "active" : "inactive", CouplerType(cnum),
                qneighbor, state, -1+qnum_i, qnum_j, qnum_u, qnum_k;
        }

        # High vertical neighbor:
        if (qnum_i < (-1+DW_M) && qnum_u == 0)
        {
            qneighbor = Linear_from_ChimInd(+1+qnum_i, qnum_j, qnum_u, qnum_k);
            cnum = Coupler_from_Qubits(qnum, qneighbor);
            state = (0 + qubit[qneighbor]) ? "active" : "inactive";
            printf "Q%04d <== C%04d|%s|%s ==> Q%04d|%s|ChimeraIndex(%d,%d,%d,%d)\n",
                qnum,
                cnum, (0 + coupler[cnum]) ? "active" : "inactive", CouplerType(cnum),
                qneighbor, state, +1+qnum_i, qnum_j, qnum_u, qnum_k;
        }
    }

    if (selector ~ /C[0-9][0-9][0-9][0-9]/)
    {
        cnum = selector;
        sub("C", "", cnum);
        cnum = 0 + cnum;

        if (cnum >= couplers)
        {
            printf "error: %s is out of range for geometry\n", selector;
            exit 1;
        }

        Qubits_from_Coupler(cnum);

        ChimInd_from_Linear(pq1);
        pq1_i = i;
        pq1_j = j;
        pq1_u = u;
        pq1_k = k;

        ChimInd_from_Linear(pq2);
        pq2_i = i;
        pq2_j = j;
        pq2_u = u;
        pq2_k = k;

        printf "Q%04d|%s|ChimeraIndex(%d,%d,%d,%d) <== C%04d|%s|%s ==> Q%04d|%s|ChimeraIndex(%d,%d,%d,%d)\n",
            pq1, (0 + qubit[pq1]) ? "active" : "inactive", pq1_i, pq1_j, pq1_u, pq1_k,
            cnum, (0 + coupler[cnum]) ? "active" : "inactive", CouplerType(cnum),
            pq2, (0 + qubit[pq2]) ? "active" : "inactive", pq2_i, pq2_j, pq2_u, pq2_k;
    }
}

function ChimInd_from_Linear(linear)
{
    k = (linear % DW_L);
    linear -= k;
    linear /= DW_L;

    u = (linear % 2);
    linear -= u;
    linear /= 2;

    j = (linear % DW_N);
    linear -= j;
    linear /= DW_N;

    i = linear;
}

function Linear_from_ChimInd(i,j,u,k,temp)
{
    return 2*DW_N*DW_L*i + 2*DW_L*j + DW_L*u + k;
}

function Coupler_from_Qubits(pqubit1, pqubit2)
{
    ChimInd_from_Linear(pqubit1);
    i1 = i;
    j1 = j;
    u1 = u;
    k1 = k;

    ChimInd_from_Linear(pqubit2);
    i2 = i;
    j2 = j;
    u2 = u;
    k2 = k;

    # Intracell coupler:
    if (i1 == i2 && j1 == j2 && u1 != u2)
    {
        if (u1 == 0) offset = k1 * DW_L + k2;
        if (u2 == 0) offset = k2 * DW_L + k1;
        return i1 * (DW_L*DW_L*DW_N) + j1 * (DW_L*DW_L) + offset;
    }

    # Horizontal Intercell coupler:
    if (i1 == i2 && ((j1 - j2 == 1) || (j1 - j2 == -1)) && u1 == 1 && u2 == 1 && k1 == k2)
    {
        if (j1 < j2) jmin = j1;
        if (j2 < j1) jmin = j2;
        return DW_L*DW_L*DW_M*DW_N + i1 * (DW_L*(DW_N - 1)) + jmin * DW_L + k1;
    }

    # Vertical Intercell coupler:
    if (((i1 - i2 == 1) || (i1 - i2 == -1)) && j1 == j2 && u1 == 0 && u2 == 0 && k1 == k2)
    {
        if (i1 < i2) imin = i1;
        if (i2 < i1) imin = i2;
        return DW_L*DW_L*DW_M*DW_N + DW_L*DW_M*(DW_N - 1) + imin * (DW_L*DW_N) + j1 * DW_L + k1;
    }

    printf "error: qubits %d and %d are not connected by a coupler\n", pqubit1, pqubit2;
    exit 1;
}

function Qubits_from_Coupler(coupler, temp_coupler, offset, ti, tj, tk0, tk1)
{
    temp_coupler = coupler;

    # Intracell coupler:
    if (coupler < DW_L*DW_L*DW_M*DW_N)
    {
        offset = coupler % (DW_L * DW_L);
        coupler -= offset;
        coupler /= (DW_L * DW_L);

        tj = coupler % DW_N;
        coupler -= tj;
        coupler /= DW_N;

        ti = coupler;

        tk1 = offset % DW_L;
        offset -= tk1;
        tk0 = offset / DW_L;

        pq1 = Linear_from_ChimInd(ti, tj, 0, tk0);
        pq2 = Linear_from_ChimInd(ti, tj, 1, tk1);

        return;
    }

    coupler = temp_coupler - DW_L*DW_L*DW_M*DW_N;

    # Horizontal Intercell coupler:
    if (coupler < DW_L*DW_M*(DW_N - 1))
    {
        tk1 = coupler % DW_L;
        coupler -= tk1;
        coupler /= DW_L;

        tj = coupler % (DW_N - 1);
        coupler -= tj;
        coupler /= (DW_N - 1);

        ti = coupler;

        pq1 = Linear_from_ChimInd(ti,   tj, 1, tk1);
        pq2 = Linear_from_ChimInd(ti, 1+tj, 1, tk1);

        return;
    }

    coupler = temp_coupler - DW_L*DW_L*DW_M*DW_N - DW_L*DW_M*(DW_N - 1);

    # Vertical Intercell coupler:
    if (coupler < DW_L*(DW_M - 1)*DW_N)
    {
        tk0 = coupler % DW_L;
        coupler -= tk0;
        coupler /= DW_L;

        tj = coupler % DW_N;
        coupler -= tj;
        coupler /= DW_N;

        ti = coupler;

        pq1 = Linear_from_ChimInd(  ti, tj, 0, tk0);
        pq2 = Linear_from_ChimInd(1+ti, tj, 0, tk0);

        return;
    }

    printf "error: coupler %d out of range\n", temp_coupler;
    exit 1;
}

function CouplerType(coupler, temp_coupler)
{
    temp_coupler = coupler;

    if (coupler < DW_L*DW_L*DW_M*DW_N) return "intracell";

    coupler -= DW_L*DW_L*DW_M*DW_N;

    if (coupler < DW_L*DW_M*(DW_N - 1)) return "horizontal";

    coupler -= DW_L*DW_M*(DW_N - 1);

    if (coupler < DW_L*(DW_M - 1)*DW_N) return "vertical";

    printf "error: coupler %d out of range\n", temp_coupler;
    exit 1;
}

' < $DW_INTERNAL__WORKSPACE/geometry
