#!/bin/bash

# Creates a QMI and then uses "dw trans" to shift the QMI
# one unit cell to the right.  The solution is then shifted
# back to the original location.

# Exit if not being run via Bash shell:
if [ ! -n "$BASH" ] ; then 
    echo Please run this script $0 with bash; exit 1; 
fi

# Get a connection to the local simulator:
dw set connection laptop

# Get a 128-qubit sampling solver from the simulator:
dw set solver c4-sw_sample

# Make new workspace directory (examples) if it does not exist:
dw mkdir -p examples

# Change to new workspace directory:
dw cd examples

# Initialize the workspace:
dw init

# Create the QMI:
dw set ring.qmi <<EOF
    Q0000 <== 5
    Q0003 <== 5
    Q0004 <== 5
    Q0007 <== 5
    C0000 <== -10
    C0003 <== -10
    C0012 <== -10
    C0015 <== -10
EOF

# Create the transform:
cat > shifter.t <<EOF
    Q0008 <== Q0000
    Q0009 <== Q0001
    Q0010 <== Q0002
    Q0011 <== Q0003
    Q0012 <== Q0004
    Q0013 <== Q0005
    Q0014 <== Q0006
    Q0015 <== Q0007
EOF

# Transform the QMI to the neighboring unit cell:
dw trans shifter.t shifted-ring.qmi ring.qmi

# Execute the QMI:
dw exec num_reads=1000 shifted-ring.qmi

# Bring the solutions back to the original unit cell:
dw trans shifter.t ring.sol shifted-ring.sol

# Display the solution:
dw get ring.sol
